/*************************************************************************
;* File:	SVGA.C    						*
;* Description: This module contains procedures to select mode and to	*
;*		select pages.  It also initializes global variables	*
;*		according to the values in the MODE.H include file.	*
;*		This module uses the VESA interrupt 10h extensions	*
;*		and is patterned after the assembly language routines	*
;*		given in the book Advanced Programmers Guide to 	*
;*		SuperVGAs by Sutty and Blair. Conversion to C by	*
;*		Lanny Murdock.						*
;* Entry Points:							*
;*		Select_Graphics   - Select a graphics mode		*
;*		Select_Page	   - Select read and write page 	*
;*		Write_Video	- Write a block of video data		*
;* Uses:								*
;*		MODE.H		   - Mode dependent constants		*
;*		VGA.H		- Graphics adapter info.		*
;*									*
;*		Following are modes for VESA BIOS boards:		*
;*	 |----- 256 colors ------|  |-- 16 colors --| 4 colors 2 colors *
;*	 640x400  640x480  800x600  800x600  1024x768 1024x768 1024x768 *
;* Mode:   100h    101h     103h      102h     104h	 N/A	 N/A	*
;*************************************************************************/

#include <dos.h>
#include <io.h>
#include  <conio.h>
#include <stdio.h>
#include <stdlib.h>

#include <at200.h>

#include "idecls.h"
#include "camsys.h"
#include "proto.h"
#include "vga.h"
#include "mode.h"
#include "chars.h"

/* Macro to make a far pointer */

#define	MK_FP(seg, off)	((void far *)\
	(((unsigned long)(seg)<<16)|(unsigned)(off)))

#define	Page_Granularity		0x10000L  /* Assume pages 64K long. */
#define Video_Pitch			640L	  /* Assume 640x480 screen  */

/* VESA driver information structure */

struct	{
	char	VESASignature[4];	/* 4 signature bytes */
	int	VESAVersion;		/* VESA nersion number */
	char far *OEMString;		/* Pointer to OEM string */
	char	Capabilities[4];	/* Video env. capabilities */
	int far *VideoModePtr;		/* Pointer to list of supported modes */
	char	Dummy[246];		/* Pad to at least 256 bytes */
	} VESA_Info;

/* Mode information structure */

struct  {
	int	ModeAttributes;		/* mode attributes*/
	char	WinAAttributes;		/* window A attributes*/
	char	WinBAttributes;		/* window B attributes*/
	int	WinGranularity;		/* window granularity*/
	int	WinSize;		/* window size*/
	int	WinASegment;		/* window A start segment*/
	int	WinBSegment;		/* window B start segment*/
	long	WinFuncPtr;		/* pointer to window function*/
	int	BytesPerScanLine;	/* bytes per scan line*/

/* Optional information (provided if bit D1 of ModeAttributes is set)*/

	int	XResolution;		/* horizontal resolution*/
	int	YResolution;		/* vertical resolution*/
	char	XCharSize;		/* character cell width*/
	char	YCharSize;		/* character cell height*/
	char	NumberOfPlanes;		/* number of memory planes*/
	char	BitsPerPixel;		/* bits per pixel*/
	char	NumberOfBanks;		/* number of banks*/
	char	MemoryModel;		/* memory model type*/
	char	BankSize;		/* bank size in kb*/  } Mode_Info;
	char	dummy[227];		/* Pad out to 256 byte structure */
	
union REGS inregs, outregs;
struct SREGS segregs;
int RW_Page;		/* Selected video page */
unsigned old_state; 	/* Space to store state of video controller */
/*
;************************************************************************
;*									*
;* Select_Graphics(HorizPtr, VertPtr, ColorsPtr)			*
;*	Initialize VGA adapter to 640x480 mode with			*
;*	256 colors.							*
;*									*
;* Entry:								*
;*	None								*
;*									*
;************************************************************************
*/

void Select_Graphics( void )
{
int x;
void far *farptr;

/* Verify presence of VESA driver */

farptr = (void far *)&VESA_Info;
inregs.x.ax = 0x4F00;
segregs.es = FP_SEG( farptr );
inregs.x.di = FP_OFF( farptr );
int86x( 0x10, &inregs, &outregs, &segregs );
if ( outregs.x.ax != 0x004F ){
   printf("VESA driver not found\n");
   exit(0);
}

/* Get information about requested mode and select it*/

farptr = (void far *)&Mode_Info;
inregs.x.ax = 0x4F01;
inregs.x.cx = GRAPHICS_MODE;
segregs.es = FP_SEG( farptr );
inregs.x.di = FP_OFF( farptr );
int86x( 0x10, &inregs, &outregs, &segregs );
if ( outregs.x.ax != 0x004F ){
   printf("This mode not supported 1\n");
   exit(0);
}

/*	; To limit number of versions for each drawing routine, only
	; 64kByte pages are supported.
	; (And even for 64kByte pages there are already 10 versions.)
	; Here a check is made that this mode is a 'simple' mode, with
	; 64k window at A000h.	It is assumed that this mode is
	; either 16-color planar or 256-color packed pixel organization.*/

if ((Mode_Info.WinSize != 64 ) || ( Mode_Info.WinASegment != 0xA000 )){
   printf("This mode not supported 2\n");
   exit( 0 );
}

inregs.x.ax = 0x4F02;
inregs.x.bx = GRAPHICS_MODE;
int86( 0x10, &inregs, &outregs );
if ( outregs.x.ax != 0x004F ){
   printf("This mode not supported 3\n");
   exit(0);
}

/*	; Reset 'last selected page'*/

	for (x = 0; x < 64; x++)
		svga_color_set(x, x, x, x);
RW_Page = 0xFF;		/* Non-existent page */

}


void svga_color_set(int color, int red, int green, int blue)
{
   outp(DAC_WRITE_PORT, color);
   outp(LOOK_UP_DATA, red);
   outp(LOOK_UP_DATA, green);
   outp(LOOK_UP_DATA, blue);
}


/*;************************************************************************
;									*
; Select_Page								*
;									*
;*************************************************************************/

void Select_Page( int page )
{
int new_page;

if ( page == RW_Page )			/*Check if already selected*/
   return;
page = page & 0x07;
RW_Page = page;				/*Save page number*/

/*	; Convert 64k page number according to board granularity*/

new_page = (int)(( 64L /(long)Mode_Info.WinGranularity ) * (long)page);

inregs.x.ax = 0x4F05;
inregs.x.dx = new_page;
inregs.x.bx = 0;
int86( 0x10, &inregs, &outregs );
}



int getvideomode( void )
{

inregs.x.ax = 0x4F03;
int86( 0x10, &inregs, &outregs );

return outregs.x.bx;
}


void setvideomode( int mode )
{

inregs.x.ax = 0x4F02;
inregs.x.bx = mode;
int86( 0x10, &inregs, &outregs );
}



/****************************************************************************
*									    *
* Write_Video( x, y, dx, dy, srcptr )					    *
*	Write a rectangular block of image data to the screen where:	    *
*		x, y are the coordinates of the upper left corner	    *
*		dx, dy are the horizontal and vertical dimensions in pixels *
*		srcptr is a pointer to the buffer containing the image data *
*									    *
*	The function Select_Graphics must be executed at least once before  *
* 	this function is called so that the structure Mode_Info contains    *
*	the correct information.					    *
*****************************************************************************/


unsigned char far *screenptr;   /* Pointer to window into video RAM. */
unsigned char far *srcptr;	/* Pointer to buffer containing image data. */

void Write_Video( int x, int y, int dx, int dy, unsigned char far *srcptr)
{
int count, page, bytes_this_page, bytes_next_page;
long offset, next_offset;

/* Write image data line by line, changing pages in video RAM when necessary. */

while ( dy-- ){
   offset = (long)(( (long)y * Video_Pitch + (long)x )%Page_Granularity);
   screenptr = MK_FP( Mode_Info.WinASegment, offset );
   next_offset = offset + (long)dx;
   
/* Make sure we're in the right page. */

   page = (int)(((long)y * Video_Pitch + (long)x)/Page_Granularity);
   Select_Page( page );

/* If a page change isn't necessary, write the whole line. */

   if ( (next_offset/Page_Granularity) == 0 ){
   	count = dx;
   	while ( count-- )
	   *screenptr++ = *srcptr++;
   	}
   	
/* If a page change is necessary, write the remaining bytes in this page,
   the rest in the next. */
   
   else {
   	bytes_next_page = (int) (next_offset - Page_Granularity);
   	bytes_this_page = dx - bytes_next_page;
   	
       if ( bytes_this_page > 0 )
           while ( bytes_this_page-- )
              *screenptr++ = *srcptr++;
        page++;
        Select_Page( page );
        screenptr = MK_FP( Mode_Info.WinASegment, 0 );  /* Offset of first byte*/
        if ( bytes_next_page > 0 )			/* in next page is 0. */
           while ( bytes_next_page-- )
              *screenptr++ = *srcptr++;
        }
   y++;          
}
}



void Draw_HLine( int x0, int x1, int y, char color )
{
int count, page, bytes_this_page, bytes_next_page, line_length;
long offset, next_offset;

/* Write image data line by line, changing pages in video RAM when necessary. */

offset = (long)(( (long)y * Video_Pitch + (long)x0 )%Page_Granularity);
screenptr = MK_FP( Mode_Info.WinASegment, offset );
line_length = x1 - x0;
next_offset = offset + (long)line_length;

/* Make sure we're in the right page. */

page = (int)(((long)y * Video_Pitch + (long)x0)/Page_Granularity);
Select_Page( page );

/* If a page change isn't necessary, write the whole line. */

if ( (next_offset/Page_Granularity) == 0 ){
   count = line_length;
   while ( count-- )
      *screenptr++ = color;
   }
   	
/* If a page change is necessary, write the remaining bytes in this page,
   the rest in the next. */
   
else {
   bytes_next_page = (int) (next_offset - Page_Granularity);
   bytes_this_page = line_length - bytes_next_page;
   	
   if ( bytes_this_page > 0 )
       while ( bytes_this_page-- )
          *screenptr++ = color;
       page++;
       Select_Page( page );
       screenptr = MK_FP( Mode_Info.WinASegment, 0 );  /* Offset of first byte*/
       if ( bytes_next_page > 0 )			/* in next page is 0. */
           while ( bytes_next_page-- )
              *screenptr++ = color;
        }
   y++;          
}





void Write_Pixel( int x, int y, char color )
{
int page;
unsigned offset;

offset = (unsigned) (( y * Video_Pitch + x ) % Page_Granularity);
page = (int)(( (long)y * (long)Video_Pitch + (long)x ) / (long)Page_Granularity);
Select_Page( page );
screenptr = MK_FP( Mode_Info.WinASegment, offset );

*screenptr = color;

}



void Draw_VLine( int x, int y0, int y1, char color )
{
int y;

for ( y = y0; y < y1; y++ )
   Write_Pixel( x, y, color );
RW_Page = 0xFF;			/* Just in case we pass a page boundary */
}





/*----------------------------------------------------------------------*/
/*   print_string
     This subroutine writes out a string to the graphics
     screen in direct write pixel by pixel mode.
*/

void print_string(int x, int y, char *string, char color)
{
     register int p = 0 ;
     char c ;

     while ((c = string[p++]) != 0)
     {
        plot_char(x,y,(c - 32) * 14,color);
        x += 8;
     }
}


/*-----------------------------------------------------------------------*/
/*   plot_char
     This subroutine plots a character directly
     to the screen in pixel by pixel mode
*/

void plot_char(int x, int y, int char_offset, char color)
{

        char c ;
	int i, j;
	
	for ( i = 0; i < 14; i++ ){
                c = char_table[char_offset++] ;
		for ( j = 0; j < 8; j++ )
		{
                        if((c & (0x80 >> j)) != 0)
				Write_Pixel( x + j, y, color );
/*			else 
				Write_Pixel( x + j, y, 0 );*/
		}
		y++;
	}
	RW_Page = 0xFF;
}



void clear( void )
{
int page = 0;
int color = 0;

unsigned count;
unsigned far * ptr;

disable_refresh();

while ( page < 5 )
   {
      ptr = ( unsigned far *)( MK_FP( Mode_Info.WinASegment, 0L ));
      Select_Page( page );
      count = 0x8000;
      while ( count-- )
         *ptr++ = color;
      page++;
   }
enable_refresh();
}


void disable_refresh( void )
{
outp( SEQUENCER_PORT, 1 );
old_state = inp( 0x3C5 );
outp( 0x3C5, old_state | 0x0020 );
}

void enable_refresh( void )
{
unsigned state;

state = ((unsigned)old_state << 8) + 1;
outpw( 0x3C4, state );
}



